from sqlalchemy.orm import Session

from qq.application import Application
from qq.initializers import ApplicationInitializer
from qq.injector import SimpleInjector
from qq.plugins.settings import TESTS_KEY
from qq.plugins.settings import SettingsInjector
from qq.plugins.types import Settings


class TransactionContextManager:
    def __init__(self, session: Session, settings: Settings):
        self.session = session
        self.settings = settings

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc, exc_tb):
        if exc_type:
            self.session.rollback()
        elif self.settings.get(TESTS_KEY, False):
            self.session.flush()
        else:
            self.session.commit()


class TransactionDecorator:
    def __init__(
        self,
        application: Application,
        key: str,
    ):
        super().__init__()
        self.key = key
        self.app_wrapper = ApplicationInitializer(application)

    def __call__(self, method):
        def wrapper(
            *args,
            session: Session = SimpleInjector(self.key),
            settings: Settings = SettingsInjector(self.key),
            **kwargs,
        ):
            with TransactionContextManager(session, settings):
                return self.app_wrapper(method)(*args, **kwargs)

        return self.app_wrapper(wrapper)



from os.path import dirname

from qq.application import Application
from qq.context import Context
from qq.injector import Injector
from qq.plugin import Plugin

TESTS_KEY = "tests"


def _import(modulepath):
    return __import__(modulepath, globals(), locals(), [""])


class PrefixedStringsDict(dict):
    def __init__(self, prefix="", module=None):
        self.prefix = prefix
        if module:
            self.set_prefix_from_module(module)

    def set_prefix(self, prefix):
        self.prefix = prefix

    def __getitem__(self, key):
        return self.prefix + super().__getitem__(key)

    def __setitem__(self, key, value):
        if not isinstance(value, str):
            raise ValueError("PrefixedStringsDict can be set only by strings!")
        return super().__setitem__(key, value)

    def set_prefix_from_module(self, module):
        self.prefix = dirname(module.__file__)


class SettingsPlugin(Plugin):
    """
    This class will generate settings for different startpoints. `modulepath`
    is a dotted path to the module with the settings startpoints. Startpoint is
    a function which will create proper settings and push them to configurator.
    """

    key = "settings"

    def __init__(self, modulepath: str):
        super().__init__()
        self.modulepath = modulepath
        self._settings = None

    def start(self, application: Application):
        startpoints_module = _import(self.modulepath)
        settings = getattr(startpoints_module, application.startpoint)
        return settings()

    def enter(self, context: Context):
        return context.globals[self.key]


class SettingsBasedPlugin(Plugin):
    def get_my_settings(self, source: (Application, Context)):
        """
        Get settings for this plugin from Application or Context
        """
        return source.globals[SettingsPlugin.key][self.key]


@Injector
def SettingsInjector(context: Context, key: str):
    return context[SettingsPlugin.key][key]



from unittest.mock import MagicMock

from pytest import fixture
from pytest import raises

from qq.errors import PluginKeyAlreadyDefinedError
from qq.errors import PluginLacksOfKeyError
from qq.errors import WrongKeyForPluginError
from qq.plugin_container import PluginContainer


class TestPluginContainer:
    @fixture
    def container(self):
        return PluginContainer()

    @fixture
    def plugin(self):
        return MagicMock(key=None)

    @fixture
    def singleton_plugin(self):
        return MagicMock(key="plugin")

    def test_calling_not_singleton_plugin(self, container, plugin):
        """
        PluginContainer should raise PluginLacksOfKeyError when plugin has no
        global .key parameter.
        """
        with raises(PluginLacksOfKeyError):
            container(plugin)

    def test_calling_singleton_plugin(self, container, singleton_plugin):
        """
        PluginContainer should accept calling plugin with global .key parameter
        defined.
        """
        container(singleton_plugin)

        assert container[singleton_plugin.key] == singleton_plugin

    def test_settings_singleton_plugin_with_wrong_key(
        self, container, singleton_plugin
    ):
        """
        PluginContainer should raise WrongKeyForPluginError when trying to set
        singleton plugin with key which is not equal to Plugin's .key parameter.
        """
        with raises(WrongKeyForPluginError):
            container["wron_name"] = singleton_plugin

    def test_settings_plugin_which_is_already_there(self, container, plugin):
        """
        PluginContainer should raise PluginKeyAlreadyDefinedError when trying
        to set plugin to a key that is already there.
        """
        container["key"] = plugin
        with raises(PluginKeyAlreadyDefinedError):
            container["key"] = plugin

    def test_setting_plugin(self, container, plugin):
        """
        PluginContainer should set the plugin to a provided key.
        """
        container["key"] = plugin
        assert container["key"] == plugin

    def test_setting_singleton_plugin(self, container, singleton_plugin):
        """
        PluginContainer should set the singleton plugin to a provided key if it
        matches the plugin's .key parameter.
        """
        container[singleton_plugin.key] = singleton_plugin
        assert container[singleton_plugin.key] == singleton_plugin



from pickle import dumps
from pickle import loads
from unittest.mock import MagicMock
from unittest.mock import sentinel

from pytest import fixture
from pytest import raises

from qq.application import AlreadyStartedError
from qq.application import Application
from qq.plugin import Plugin


class ExamplePlugin(Plugin):
    def enter(self, context):
        context["example"] = sentinel.example


class ExampleApplication(Application):
    def create_plugins(self):
        super().create_plugins()
        self.plugins["plugin1"] = MagicMock(key=None)
        self.plugins["plugin2"] = MagicMock(key=None)
        self.plugins["plugin3"] = ExamplePlugin()


class TestApplication:
    @fixture
    def app(self):
        return ExampleApplication()

    def test_start(self, app):
        """
        .start should append plugins and init them. Also proper flags should be
        set.
        """
        app.start("wsgi", wsgi=1)

        plugin1 = app.plugins["plugin1"]
        plugin2 = app.plugins["plugin2"]

        assert app.extra == {"wsgi": 1}
        assert app.globals == {
            "plugin1": plugin1.start.return_value,
            "plugin2": plugin2.start.return_value,
            "plugin3": None,
        }
        assert app.startpoint == "wsgi"
        assert app.is_started

        plugin1.start.assert_called_once_with(app)
        plugin2.start.assert_called_once_with(app)

    def test_start_when_started(self, app):
        """
        .start should do nothing if app already started
        """
        app.is_started = True

        with raises(AlreadyStartedError):
            assert app.start("wsgi", wsgi=1)

        assert app.globals == {}
        assert app.startpoint is None
        assert app.plugins == {}

    def test_pickling(self, app):
        pickle = dumps(app)
        unpickle = loads(pickle)

        assert app.is_started == unpickle.is_started
        assert app.startpoint == unpickle.startpoint
        assert app.extra == unpickle.extra



from copy import copy
from functools import wraps
from inspect import BoundArguments
from inspect import iscoroutinefunction
from inspect import signature
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple

from qq.application import Application
from qq.types import BaseInjector

QQ_PARAMETER = "_qq_application"


class DefaultEmptyVar:
    pass


class ArgumentInitializer:
    """
    ArgumentInitializer makes sure that the injecotr will be started and
    ended after function is finished or raised an error.
    """

    def __init__(self, initializer, fun, args, kwargs):
        self.initializer = initializer
        self.fun = fun
        self.args = args
        self.kwargs = kwargs
        self.injectors = None

    def __enter__(self):
        kwargs = copy(self.kwargs)
        self.injectors = list(get_injectors(self.fun, self.args, self.kwargs))
        for name, injector in self.injectors:
            kwargs[name] = injector.start(self.initializer.application)

        initializers = list(get_initializers(self.fun, self.args, self.kwargs))
        for name, initializer in initializers:
            kwargs[name] = self.initializer.wrapper(initializer)
        return kwargs

    def __exit__(self, exc_type, exc_value, traceback):
        for name, injector in reversed(self.injectors):
            injector.end()


class ApplicationInitializer:
    def __init__(self, application: Application):
        self.application = application

    def __call__(self, fun: Callable) -> Callable:
        """
        If function has injectors in the defaults of arguments, then init them.
        """
        return wraps(fun)(self.wrapper(fun))

    def wrapper(self, fun: Callable):
        if getattr(fun, QQ_PARAMETER, None):
            setattr(fun, QQ_PARAMETER, self.application)
            return fun

        @wraps(fun)
        def qqfunction(*args, **kwargs):
            icm = ArgumentInitializer(self, fun, args, kwargs)
            with icm as kwargs:
                return fun(*args, **kwargs)

        @wraps(fun)
        async def qqcoroutine(*args, **kwargs):
            icm = ArgumentInitializer(self, fun, args, kwargs)
            with icm as kwargs:
                return await fun(*args, **kwargs)

        setattr(qqfunction, QQ_PARAMETER, self.application)
        setattr(qqcoroutine, QQ_PARAMETER, self.application)
        return qqcoroutine if iscoroutinefunction(fun) else qqfunction


def is_injector(parameter: Any, name: str, bound_args: BoundArguments):
    """
    Is parameter an injector and the value is not provided in the call.
    """
    value = (
        bound_args.arguments[name]
        if name in bound_args.arguments
        else parameter._default
    )
    return isinstance(value, BaseInjector)


def is_initializer(parameter: Any, name: str, bound_args: BoundArguments):
    """
    Is parameter an injector and the value is not provided in the call.
    """
    qq_application = getattr(parameter._default, QQ_PARAMETER, DefaultEmptyVar)
    is_injected = qq_application != DefaultEmptyVar
    return is_injected and name not in bound_args.arguments


def get_injectors(
    method: Callable,
    args: List,
    kwargs: Dict,
) -> Iterable[Tuple[str, Callable]]:
    """
    Generate list of iterators that was not overwrited by function's call args.
    """
    sig = signature(method)
    bound_args = sig.bind_partial(*args, **kwargs)

    for name, parameter in sig._parameters.items():
        if is_injector(parameter, name, bound_args):
            yield name, parameter._default


def get_initializers(
    method: Callable,
    args: List,
    kwargs: Dict,
) -> Iterable[Tuple[str, Callable]]:
    """ """
    sig = signature(method)
    bound_args = sig.bind_partial(*args, **kwargs)

    for name, parameter in sig._parameters.items():
        if is_initializer(parameter, name, bound_args):
            yield name, parameter._default



from qq.application import Application
from qq.context import ApplicationNotStartedError
from qq.context import Context
from qq.initializers import ApplicationInitializer
from qq.injector import ContextManagerInjector
from qq.injector import Injector
from qq.injector import SimpleInjector
from qq.plugin import Plugin
